#include "modest_config.h"

// myhtml
#include "third_party/modest/source/myhtml/tree.c"
#include "third_party/modest/source/myhtml/data_process.c"
#include "third_party/modest/source/myhtml/mynamespace.c"
#include "third_party/modest/source/myhtml/tokenizer.c"
#include "third_party/modest/source/myhtml/tokenizer_doctype.c"
#include "third_party/modest/source/myhtml/tokenizer_end.c"
#include "third_party/modest/source/myhtml/tokenizer_script.c"
#include "third_party/modest/source/myhtml/tag_init.c"
#include "third_party/modest/source/myhtml/rules.c"
#include "third_party/modest/source/myhtml/callback.c"
#include "third_party/modest/source/myhtml/serialization.c"
#include "third_party/modest/source/myhtml/myhtml.c"
#include "third_party/modest/source/myhtml/mystring.c"
#include "third_party/modest/source/myhtml/tag.c"
#include "third_party/modest/source/myhtml/token.c"
#include "third_party/modest/source/myhtml/stream.c"
#include "third_party/modest/source/myhtml/charef.c"
#include "third_party/modest/source/myhtml/parser.c"
